from django.core.exceptions import ImproperlyConfigured

try:
    import psycopg2

    HAS_PSYCOPG2 = True
except ImportError:
    psycopg2 = None
    HAS_PSYCOPG2 = False

try:
    from django.contrib.gis.geos.libgeos import geos_version_info

    HAS_GEOS = geos_version_info()["version"] >= "3.3.0"
except (ImportError, OSError, ImproperlyConfigured):
    HAS_GEOS = False



from typing import Any, AnyStr, Callable, Dict, List, Tuple, Type, TypeVar, Union

from django.db import models

T = TypeVar("T", bound=models.Model)
Seed = Union[AnyStr, bytearray, int]
FieldMap = Dict[str, Any]
Lookup = Dict[str, Any]
SaveHooks = List[Callable[[T], None]]
LazySaveHooks = List[Callable[[models.Model], None]]
Built = Tuple[T, Dict[str, List[models.Model]]]
LazyBuilt = Tuple[models.Model, Dict[str, List[models.Model]]]



from functools import partial
from typing import Any, Callable, Generic, List
from typing import Optional as Opt
from typing import overload

from .types import Built, FieldMap, SaveHooks, Seed, T


class Blueprint(Generic[T]):
    def __init__(self, model, fields=None, pre_save=None, post_save=None, seed=None):
        # type: (T, Opt[FieldMap], Opt[SaveHooks], Opt[SaveHooks], Opt[Seed]) -> None
        from .faker_factory import factory

        self.factory = factory

        self._model = model
        self._fields = fields or {}
        self.seed = seed
        self.pre_save = pre_save
        self.post_save = post_save

        self.pk = -1

    def fields(self, **kwargs):
        # type: (**Any) -> "Blueprint"
        self._fields = kwargs
        return self

    def make_one(self, fields=None, pre_save=None, post_save=None, seed=None, iteration=None):
        # type: (Opt[FieldMap], Opt[SaveHooks], Opt[SaveHooks], Opt[Seed], Opt[int]) -> T
        _fields = self._fields.copy()
        if fields:
            _fields.update(fields)
        if seed is None:
            seed = self.seed

        if pre_save is None:
            pre_save = self.pre_save

        if post_save is None:
            post_save = self.post_save

        return self.factory.make_one(
            self._model, _fields, pre_save, post_save, seed, iteration
        )

    @overload
    def make(self, fields, pre_save, post_save, seed, quantity):  # pragma: no cover
        # type: (Opt[FieldMap], Opt[SaveHooks], Opt[SaveHooks], Opt[Seed], None) -> T
        pass

    @overload
    def make(self, fields, pre_save, post_save, seed, quantity):  # pragma: no cover
        # type: (Opt[FieldMap], Opt[SaveHooks], Opt[SaveHooks], Opt[Seed], int) -> List[T]
        pass

    def make(
        self, fields=None, pre_save=None, post_save=None, seed=None, quantity=None
    ):
        _fields = self._fields.copy()
        if fields:
            _fields.update(fields)
        if seed is None:
            seed = self.seed

        if pre_save is None:
            pre_save = self.pre_save

        if post_save is None:
            post_save = self.post_save

        return self.factory.make(
            self._model, _fields, pre_save, post_save, seed, quantity
        )

    @overload
    def build(self, fields, pre_save, seed, quantity, make_fks):  # pragma: no cover
        # type: (Opt[FieldMap], Opt[SaveHooks], Opt[Seed], None, bool) -> Built
        pass

    @overload
    def build(self, fields, pre_save, seed, quantity, make_fks):  # pragma: no cover
        # type: (Opt[FieldMap], Opt[SaveHooks], Opt[Seed], int, bool) -> List[Built]
        pass

    def build(self, fields=None, pre_save=None, seed=None, quantity=None, make_fks=False):
        _fields = self._fields.copy()
        if fields:
            _fields.update(fields)
        if seed is None:
            seed = self.seed

        if pre_save is None:
            pre_save = self.pre_save

        return self.factory.build(
            self._model, _fields, pre_save, seed, quantity, make_fks
        )

    @overload
    def m(self, pre_save, post_save, seed, quantity):  # pragma: no cover
        # type: (Opt[SaveHooks], Opt[SaveHooks], Opt[Seed], None) -> Callable[..., T]
        pass

    @overload
    def m(self, pre_save, post_save, seed, quantity):  # pragma: no cover
        # type: (Opt[SaveHooks], Opt[SaveHooks], Opt[Seed], int) -> Callable[..., List[T]]
        pass

    def m(self, pre_save=None, post_save=None, seed=None, quantity=None):
        make = partial(
            self.make,
            pre_save=pre_save,
            post_save=post_save,
            seed=seed,
            quantity=quantity,
        )

        def fn(**kwargs):
            return make(fields=kwargs)

        return fn

    @overload
    def b(self, pre_save, seed, quantity, make_fks):  # pragma: no cover
        # type: (Opt[SaveHooks], Opt[Seed], None, bool) -> Callable[..., Built]
        pass

    @overload
    def b(self, pre_save, seed, quantity, make_fks):  # pragma: no cover
        # type: (Opt[SaveHooks], Opt[Seed], int, bool) -> Callable[..., List[Built]]
        pass

    def b(self, pre_save=None, seed=None, quantity=None, make_fks=False):
        build = partial(self.build, pre_save=pre_save, seed=seed, quantity=quantity, make_fks=make_fks)

        def fn(**kwargs):
            return build(fields=kwargs)

        return fn



from django.utils import text, timezone

from faker.generator import random

from .compat import HAS_GEOS, HAS_PSYCOPG2


def comma_sep_integers(faker, field, *args, **kwargs):
    return ",".join([faker.random_int() for _ in range(10)])


def decimal(faker, field, *args, **kwargs):
    right_digits = field.decimal_places
    left_digits = field.max_digits - right_digits
    return faker.pydecimal(
        left_digits=left_digits, right_digits=right_digits, positive=True
    )


def slug(faker, field, count, *args, **kwargs):
    return text.slugify("-".join(faker.words(nb=count)))[: field.max_length]


if HAS_GEOS:
    from django.contrib.gis import gdal, geos

    def point(faker, field, srid):
        lat = random.uniform(-180.0, 180.0)
        lng = random.uniform(-90, 90)

        if field.dim == 2:
            return geos.Point(lat, lng, srid=srid)
        else:
            alt = random.uniform(-4000.0, 9000.0)
            return geos.Point(lat, lng, alt, srid=srid)

    def linestring(faker, field, srid, *args, **kwargs):
        point_count = faker.random_int(min=2, max=10)

        points = [point(faker, field, srid) for _ in range(point_count)]
        return geos.LineString(*points)

    def linearring(faker, field, srid, *args, **kwargs):
        point_0 = point(faker, field, srid)
        point_1 = point(faker, field, srid)
        point_2 = point(faker, field, srid)
        point_3 = geos.Point(point_0.x, point_0.y)
        points = [point_0, point_1, point_2, point_3]
        return geos.LinearRing(*points)

    def polygon(faker, field, srid, *args, **kwargs):
        ring = linearring(faker, field, srid)
        return geos.Polygon(ring)

    def collection(faker, field, srid, element_func, geometry_class, *args, **kwargs):
        element_count = faker.random_int(min=1, max=10)
        elements = [element_func(faker, field, srid) for _ in range(element_count)]

        return geometry_class(*elements)

    def multipoint(faker, field, srid, *args, **kwargs):
        return collection(faker, field, srid, point, geos.MultiPoint)

    def multilinestring(faker, field, srid, *args, **kwargs):
        return collection(faker, field, srid, linestring, geos.MultiLineString)

    def multipolygon(faker, field, srid, *args, **kwargs):
        return collection(faker, field, srid, polygon, geos.MultiPolygon)

    def geometrycollection(faker, field, srid, *args, **kwargs):
        single_point = point(faker, field, srid)
        points = collection(faker, field, srid, point, geos.MultiPoint)
        geometries = [single_point] + points
        return geos.GeometryCollection(*geometries)

    def gdal_raster(faker, field, srid, *args, **kwargs):
        scale = faker.pyfloat(positive=True)
        return gdal.GDALRaster(
            {
                "width": faker.random_int(min=1, max=1024),
                "height": faker.random_int(min=1, max=1024),
                "name": faker.word(),
                "srid": srid,
                "scale": [scale, -scale],
                "bands": [{"data": range(faker.random_int(min=1, max=1024))}],
            }
        )


if HAS_PSYCOPG2:
    from psycopg2.extras import DateRange, DateTimeTZRange, NumericRange

    def array(faker, field, *args, **kwargs):
        from .field_mappings import mappings_types
        from .values import Evaluator

        if field.size:
            size = field.size
        else:
            size = 10
        element_count = faker.random_int(min=1, max=size)

        evaluator = Evaluator(faker=faker, factory=None, iteration=None)

        fake = mappings_types[field.base_field.__class__]
        return [evaluator.evaluate_fake(fake, field) for _ in range(element_count)]

    def integerrange(faker, field, min, max, *args, **kwargs):
        lower = faker.random_int(min=min, max=max - 1)
        upper = faker.random_int(min=lower, max=max)
        return NumericRange(lower, upper)

    def floatrange(faker, field, *args, **kwargs):
        lower = random.uniform(-2147483647, 2147483646)
        upper = random.uniform(lower, 2147483647)
        return NumericRange(lower, upper)

    def datetimerange(faker, field, *args, **kwargs):
        lower = timezone.make_aware(faker.date_time())
        upper = timezone.make_aware(faker.date_time_between_dates(datetime_start=lower))
        return DateTimeTZRange(lower, upper)

    def daterange(faker, field, *args, **kwargs):
        lower = faker.date_time().date()
        upper = faker.date_time_between_dates(datetime_start=lower).date()
        return DateRange(lower, upper)

    def random_dict(faker, field, *args, **kwargs):
        return faker.pydict(10, True, int, str)



from typing import Any


class Lazy(object):
    def __init__(self, name, *args, **kwargs):
        # type (str, *Any, **Any) -> None
        self.name = name
        self.args = args
        self.kwargs = kwargs
        super(Lazy, self).__init__()



import random

from functools import partial
from typing import Any, Callable, Dict, Generic, List
from typing import Optional as Opt
from typing import Tuple, Union, overload

from django.conf import settings
from django.contrib.auth import get_user_model
from django.contrib.contenttypes.fields import GenericForeignKey, GenericRelation
from django.db import models
from django.db.models.fields import NOT_PROVIDED
from django.forms.models import model_to_dict

from faker import Factory as FakerFactory

from . import rels
from .blueprint import Blueprint
from .exceptions import ForeignKeyError
from .lazy import Lazy
from .types import Built, FieldMap, LazyBuilt, LazySaveHooks, Lookup, SaveHooks, Seed, T
from .utils import get_model, get_model_fields, language_to_locale, set_related
from .values import Evaluator

user_model = get_user_model()

locale = language_to_locale(settings.LANGUAGE_CODE)


class Empty(object):
    pass


fks_cache = {}  # type: Dict[str, Any]


class Factory(Generic[T]):
    def __init__(self, fake=None):
        # type: (Opt[Factory]) -> None
        self.fake = fake or FakerFactory.create(locale)

    def _serialize_instance(self, instance):
        # type: (models.Model) -> FieldMap
        model_fields = dict(get_model_fields(instance))
        attrs = {}  # type: FieldMap
        for k, v in model_to_dict(instance).items():
            if k == instance._meta.pk.name:  # type: ignore
                continue

            if isinstance(v, (list, models.QuerySet)):
                continue

            if isinstance(model_fields[k], models.ForeignKey) and not isinstance(
                v, models.Model
            ):
                attrs[k + "_id"] = v
                continue

            attrs[k] = v

        return attrs

    def seed(self, seed, set_global=False):
        # type: (Seed, bool) -> None
        self.fake.seed(seed)
        if set_global:
            random.seed(seed)

    def blueprint(self, model, *args, **kwargs):
        # type: (Union[str, models.Model], *Any, **Any) -> Blueprint
        return Blueprint(get_model(model), *args, **kwargs)

    @overload
    def build_one(self, model, fields, pre_save, seed, make_fks, iteration):  # pragma: no cover
        # type: (str, Opt[FieldMap], Opt[LazySaveHooks], Opt[Seed], bool, Opt[int]) -> LazyBuilt
        pass

    @overload
    def build_one(self, model, fields, pre_save, seed, make_fks, iteration):  # pragma: no cover
        # type: (T, Opt[FieldMap], Opt[SaveHooks], Opt[Seed], bool, Opt[int]) -> Built
        pass

    def build_one(
        self,
        model,
        fields=None,
        pre_save=None,
        seed=None,
        make_fks=False,
        iteration=None
    ):
        if fields is None:
            fields = {}

        if pre_save is None:
            pre_save = []

        if seed:
            fake = FakerFactory.create(locale)
            fake.seed(seed)
        else:
            fake = self.fake

        evaluator = Evaluator(fake, factory=self, iteration=iteration)

        model = get_model(model)
        instance = model()
        m2ms = {}
        lazies = []

        model_fields = get_model_fields(model)
        for _field_name, model_field in model_fields:
            value = Empty
            field_name = _field_name

            if _field_name.endswith("_id") and model_field.is_relation:
                continue

            if isinstance(model_field, models.AutoField):
                continue

            if isinstance(model_field, (GenericForeignKey, GenericRelation)):
                continue

            if field_name not in fields and (
                model_field.null or model_field.default != NOT_PROVIDED
            ):
                continue

            if field_name not in fields and isinstance(
                model_field, models.ManyToManyField
            ):
                continue

            value = fields.get(field_name, Empty)
            if isinstance(model_field, models.ForeignKey):
                if value == Empty:
                    value = fields.get(field_name + "_id", Empty)

                if value == rels.SELECT:
                    model = model_field.related_model
                    qs = model.objects.all()
                    cache_key = model

                    value = fks_cache.get(cache_key, Empty)

                    if value == Empty:
                        try:
                            value = qs[0]
                        except IndexError:
                            value = evaluator.fake_value(model, model_field)

                        fks_cache[cache_key] = value

                if not make_fks and ((value == Empty) or (value and value.pk is None)):
                    raise ForeignKeyError(
                        "Field {} is a required ForeignKey, but the related {}.{} model"
                        " doesn't have the necessary primary key.".format(
                            field_name,
                            model_field.related_model._meta.app_label,
                            model_field.related_model._meta.model_name,
                        )
                    )

                field_name += "_id"

            if value != Empty:
                value = evaluator.evaluate(value)
            else:
                if model_field.choices:
                    value = fake.random_element(model_field.choices)[0]
                else:
                    value = evaluator.fake_value(model, model_field)

            if isinstance(value, Lazy):
                lazies.append((field_name, value))
                continue

            if isinstance(model_field, models.ForeignKey):
                value = value.pk if hasattr(value, "pk") else value

            if isinstance(model_field, models.ManyToManyField):
                m2ms[field_name] = value
            # special case for user passwords
            if model == user_model and field_name == "password":
                instance.set_password(value)
            else:
                if field_name not in m2ms:
                    setattr(instance, field_name, value)

        for field_name, lazy in lazies:
            value = getattr(instance, lazy.name)
            if callable(value):
                value = value(*lazy.args, **lazy.kwargs)
            setattr(instance, field_name, value)

        for func in pre_save:
            func(instance)

        return instance, m2ms

    @overload
    def build(self, model, fields, pre_save, seed, quantity, make_fks):  # pragma: no cover
        # type: (str, Opt[FieldMap], Opt[LazySaveHooks], Opt[Seed], None, bool) -> LazyBuilt
        pass

    @overload
    def build(self, model, fields, pre_save, seed, quantity, make_fks):  # pragma: no cover
        # type: (T, Opt[FieldMap], Opt[SaveHooks], Opt[Seed], None, bool) -> Built
        pass

    @overload
    def build(self, model, fields, pre_save, seed, quantity, make_fks):  # pragma: no cover
        # type: (str, Opt[FieldMap], Opt[LazySaveHooks], Opt[Seed], int, bool) -> List[LazyBuilt]
        pass

    @overload
    def build(self, model, fields, pre_save, seed, quantity, make_fks):  # pragma: no cover
        # type: (T, Opt[FieldMap], Opt[SaveHooks], Opt[Seed], int, bool) -> List[Built]
        pass

    def build(
        self,
        model,
        fields=None,
        pre_save=None,
        seed=None,
        quantity=None,
        make_fks=False
    ):
        if fields is None:
            fields = {}

        if quantity is None:
            return self.build_one(model, fields, pre_save, seed, make_fks)[0]
        else:
            return [
                self.build_one(model, fields, pre_save, seed, make_fks, i)[0]
                for i in range(quantity)
            ]

    @overload
    def make_one(self, model, fields, pre_save, post_save, seed, iteration):  # pragma: no cover
        # type: (str, Opt[FieldMap], Opt[LazySaveHooks], Opt[LazySaveHooks], Opt[Seed], Opt[int]) -> models.Model
        pass

    @overload
    def make_one(self, model, fields, pre_save, post_save, seed, iteration):  # pragma: no cover
        # type: (T, Opt[FieldMap], Opt[SaveHooks], Opt[SaveHooks], Opt[Seed], Opt[int]) -> T
        pass

    def make_one(
        self,
        model,
        fields=None,
        pre_save=None,
        post_save=None,
        seed=None,
        iteration=None
    ):
        if fields is None:
            fields = {}

        if post_save is None:
            post_save = []

        instance, m2ms = self.build_one(
            model, fields, pre_save, seed, make_fks=True, iteration=iteration
        )

        # Sometimes the model's field for the primary key as a default, which
        # means ``instance.pk`` is already set. We pass ``force_insert`` as a
        # way to tell downstream code that this is a new model.
        instance.save(force_insert=True)

        for field, relateds in m2ms.items():
            set_related(instance, field, relateds)

        for func in post_save:
            func(instance)
        return instance

    @overload
    def get_or_make(self, model, lookup, fields, pre_save, post_save, seed):  # pragma: no cover
        # type: (str, Opt[Lookup], Opt[FieldMap], Opt[LazySaveHooks], Opt[LazySaveHooks], Opt[Seed]) -> Tuple[models.Model, bool]
        pass

    @overload
    def get_or_make(self, model, lookup, fields, pre_save, post_save, seed):  # pragma: no cover
        # type: (T, Opt[Lookup], Opt[FieldMap], Opt[SaveHooks], Opt[SaveHooks], Opt[Seed]) -> Tuple[T, bool]
        pass

    def get_or_make(
        self, model, lookup=None, fields=None, pre_save=None, post_save=None, seed=None
    ):
        if lookup is None:
            lookup = {}
        if fields is None:
            fields = {}
        if post_save is None:
            post_save = []

        instance, m2ms = self.build_one(model, fields, pre_save, seed, make_fks=True)

        attrs = self._serialize_instance(instance)
        for k in lookup:
            attrs.pop(k, None)
        instance, created = get_model(model).objects.get_or_create(
            defaults=attrs, **lookup
        )

        for field, relateds in m2ms.items():
            set_related(instance, field, relateds)

        for func in post_save:
            func(instance)
        return instance, created

    @overload
    def g_m(self, model, lookup, pre_save, post_save, seed):  # pragma: no cover
        # type: (str, Opt[Lookup], Opt[LazySaveHooks], Opt[LazySaveHooks], Opt[Seed]) -> Callable[..., models.Model]
        pass

    @overload
    def g_m(self, model, lookup, pre_save, post_save, seed):  # pragma: no cover
        # type: (T, Opt[Lookup], Opt[SaveHooks], Opt[SaveHooks], Opt[Seed]) -> Callable[..., T]
        pass

    def g_m(self, model, lookup=None, pre_save=None, post_save=None, seed=None):
        build = partial(
            self.get_or_make,
            model=model,
            lookup=lookup,
            pre_save=pre_save,
            post_save=post_save,
            seed=seed,
        )

        def fn(**kwargs):
            return build(fields=kwargs)

        return fn

    @overload
    def update_or_make(self, model, lookup, fields, pre_save, post_save, seed):  # pragma: no cover
        # type: (str, Opt[Lookup], Opt[FieldMap], Opt[LazySaveHooks], Opt[LazySaveHooks], Opt[Seed]) -> Tuple[models.Model, bool]
        pass

    @overload
    def update_or_make(self, model, lookup, fields, pre_save, post_save, seed):  # pragma: no cover
        # type: (T, Opt[Lookup], Opt[FieldMap], Opt[SaveHooks], Opt[SaveHooks], Opt[Seed]) -> Tuple[T, bool]
        pass

    def update_or_make(
        self, model, lookup=None, fields=None, pre_save=None, post_save=None, seed=None
    ):
        if lookup is None:
            lookup = {}
        if fields is None:
            fields = {}
        if post_save is None:
            post_save = []

        model_class = get_model(model)

        try:
            instance = model_class.objects.get(**lookup)
        except model_class.DoesNotExist:
            created = True
            params = {k: v for k, v in lookup.items() if "__" not in k}
            params.update(fields)
            instance = self.make(model, params, pre_save, post_save, seed)
        else:
            created = False
            for k, v in fields.items():
                # special case for user passwords
                if model_class == user_model and k == "password":
                    instance.set_password(v)
                else:
                    setattr(instance, k, v)
            instance.save()

            for func in post_save:
                func(instance)

        return instance, created

    @overload
    def u_m(self, model, lookup, pre_save, post_save, seed):  # pragma: no cover
        # type: (str, Opt[Lookup], Opt[LazySaveHooks], Opt[LazySaveHooks], Opt[Seed]) -> Callable[..., models.Model]
        pass

    @overload
    def u_m(self, model, lookup, pre_save, post_save, seed):  # pragma: no cover
        # type: (T, Opt[Lookup], Opt[SaveHooks], Opt[SaveHooks], Opt[Seed]) -> Callable[..., T]
        pass

    def u_m(self, model, lookup=None, pre_save=None, post_save=None, seed=None):
        build = partial(
            self.update_or_make,
            model=model,
            lookup=lookup,
            pre_save=pre_save,
            post_save=post_save,
            seed=seed,
        )

        def fn(**kwargs):
            return build(fields=kwargs)

        return fn

    @overload
    def make(self, model, fields, pre_save, post_save, seed, quantity):  # pragma: no cover
        # type: (str, Opt[FieldMap], Opt[LazySaveHooks], Opt[LazySaveHooks], Opt[Seed], None) -> models.Model
        pass

    @overload
    def make(self, model, fields, pre_save, post_save, seed, quantity):  # pragma: no cover
        # type: (T, Opt[FieldMap], Opt[SaveHooks], Opt[SaveHooks], Opt[Seed], None) -> T
        pass

    @overload
    def make(self, model, fields, pre_save, post_save, seed, quantity):  # pragma: no cover
        # type: (str, Opt[FieldMap], Opt[LazySaveHooks], Opt[LazySaveHooks], Opt[Seed], int) -> List[models.Model]
        pass

    @overload
    def make(self, model, fields, pre_save, post_save, seed, quantity):  # pragma: no cover
        # type: (T, Opt[FieldMap], Opt[SaveHooks], Opt[SaveHooks], Opt[Seed], int) -> List[T]
        pass

    def make(
        self,
        model,
        fields=None,
        pre_save=None,
        post_save=None,
        seed=None,
        quantity=None
    ):
        if fields is None:
            fields = {}
        if quantity is None:
            return self.make_one(model, fields, pre_save, post_save, seed)
        else:
            return [
                self.make_one(model, fields, pre_save, post_save, seed, i)
                for i in range(quantity)
            ]

    @overload
    def m(self, model, pre_save, post_save, seed, quantity):  # pragma: no cover
        # type: (str, Opt[LazySaveHooks], Opt[LazySaveHooks], Opt[Seed], None) -> Callable[..., models.Model]
        pass

    @overload
    def m(self, model, pre_save, post_save, seed, quantity):  # pragma: no cover
        # type: (T, Opt[SaveHooks], Opt[SaveHooks], Opt[Seed], None) -> Callable[..., T]
        pass

    @overload
    def m(self, model, pre_save, post_save, seed, quantity):  # pragma: no cover
        # type: (str, Opt[LazySaveHooks], Opt[LazySaveHooks], Opt[Seed], int) -> Callable[..., List[models.Model]]
        pass

    @overload
    def m(self, model, pre_save, post_save, seed, quantity):  # pragma: no cover
        # type: (T, Opt[SaveHooks], Opt[SaveHooks], Opt[Seed], int) -> Callable[..., List[T]]
        pass

    def m(self, model, pre_save=None, post_save=None, seed=None, quantity=None):
        make = partial(
            self.make,
            model=model,
            pre_save=pre_save,
            post_save=post_save,
            seed=seed,
            quantity=quantity,
        )

        def fn(**kwargs):
            return make(fields=kwargs)

        return fn

    @overload
    def b(self, model, pre_save, seed, quantity, make_fks):  # pragma: no cover
        # type: (str, Opt[LazySaveHooks], Opt[Seed], None, bool) -> Callable[..., LazyBuilt]
        pass

    @overload
    def b(self, model, pre_save, seed, quantity, make_fks):  # pragma: no cover
        # type: (T, Opt[SaveHooks], Opt[Seed], None, bool) -> Callable[..., Built]
        pass

    @overload
    def b(self, model, pre_save, seed, quantity, make_fks):  # pragma: no cover
        # type: (str, Opt[LazySaveHooks], Opt[Seed], int, bool) -> Callable[..., List[LazyBuilt]]
        pass

    @overload
    def b(self, model, pre_save, seed, quantity, make_fks):  # pragma: no cover
        # type: (T, Opt[SaveHooks], Opt[Seed], int, bool) -> Callable[..., List[Built]]
        pass

    def b(self, model, pre_save=None, seed=None, quantity=None, make_fks=False):
        build = partial(
            self.build,
            model=model,
            pre_save=pre_save,
            seed=seed,
            quantity=quantity,
            make_fks=make_fks,
        )

        def fn(**kwargs):
            return build(fields=kwargs)

        return fn


factory = Factory()  # type: Factory



import sys

from collections import OrderedDict

from django.conf import settings
from django.contrib.postgres import fields as pg_fields
from django.db import models
from django.utils import timezone

from . import fakes
from .compat import HAS_GEOS, HAS_PSYCOPG2

STRING_FIELDS = (
    models.CharField,
    models.TextField,
    pg_fields.CICharField,
    pg_fields.CITextField,
)

"""
This module maps fields to functions generating values.

It first tries by looking at the field's class, then falls back to some
special-cased names.

Values are 3-tuples composed of ``(<function>, <args>, <kwargs>)``.

When ``<function>`` is a string, it's assumed to be a faker provider. Whenever
``faker`` doesn't provide a suitable provider, we ship our own function. They
are defined in ``django_fakery.fakes``.
"""

TZINFO = timezone.get_current_timezone() if settings.USE_TZ else None


mappings_types = OrderedDict(
    [
        (
            models.BigIntegerField,
            ("random_int", [], {"min": -sys.maxsize, "max": sys.maxsize}),
        ),
        (models.BinaryField, ("binary", [1024], {})),
        (models.BooleanField, ("pybool", [], {})),
        (models.DateTimeField, ("date_time", [], {"tzinfo": TZINFO})),
        # ``DateField`` must come after ``DateTimeField`` because it's its superclass
        (models.DateField, (lambda faker, field: faker.date_time().date(), [], {})),
        (models.DecimalField, (fakes.decimal, [], {})),
        (models.EmailField, ("email", [], {})),
        (models.FileField, ("file_name", [], {})),
        (models.FilePathField, ("file_name", [], {})),
        (models.FloatField, ("pyfloat", [], {})),
        (models.ImageField, ("file_name", [], {"extension": "jpg"})),
        (models.IntegerField, ("pyint", [], {})),
        (models.IPAddressField, ("ipv4", [], {})),
        (models.GenericIPAddressField, ("ipv4", [], {})),
        (models.PositiveIntegerField, ("random_int", [], {"max": 2147483647})),
        (models.PositiveSmallIntegerField, ("random_int", [], {"max": 32767})),
        (models.SlugField, (fakes.slug, [3], {})),
        (models.SmallIntegerField, ("random_int", [], {"min": -32768, "max": 32767})),
        (
            models.TextField,
            (
                lambda faker, field: field.unique
                and faker.pystr(max_chars=2700)
                or faker.paragraph(),
                [],
                {},
            ),
        ),
        (
            models.TimeField,
            (lambda faker, field: faker.date_time(tzinfo=TZINFO).time(), [], {}),
        ),
        (models.URLField, ("url", [], {})),
        (
            models.CharField,
            (
                lambda faker, field: field.unique
                and faker.pystr(max_chars=field.max_length)
                or faker.word()[: field.max_length],
                [],
                {},
            ),
        ),
        (models.DurationField, ("time_delta", [], {})),
        (models.UUIDField, ("uuid4", [], {})),
    ]
)


if HAS_GEOS:
    from django.contrib.gis.db import models as geo_models

    mappings_types.update(
        {
            geo_models.PointField: (fakes.point, (), {"srid": 4326}),
            geo_models.LineStringField: (fakes.linestring, (), {"srid": 4326}),
            geo_models.PolygonField: (fakes.polygon, (), {"srid": 4326}),
            geo_models.MultiPointField: (fakes.multipoint, (), {"srid": 4326}),
            geo_models.MultiLineStringField: (
                fakes.multilinestring,
                (),
                {"srid": 4326},
            ),
            geo_models.MultiPolygonField: (fakes.multipolygon, (), {"srid": 4326}),
            geo_models.GeometryCollectionField: (
                fakes.geometrycollection,
                (),
                {"srid": 4326},
            ),
            geo_models.RasterField: (fakes.gdal_raster, [], {"srid": 4326}),
        }
    )


if HAS_PSYCOPG2:
    mappings_types.update(
        {
            pg_fields.CICharField: mappings_types[models.CharField],
            pg_fields.CIEmailField: mappings_types[models.EmailField],
            pg_fields.CITextField: mappings_types[models.TextField],
            pg_fields.ArrayField: (fakes.array, [], {}),
            pg_fields.HStoreField: ("pydict", [10, True, "str"], {}),
            pg_fields.IntegerRangeField: (
                fakes.integerrange,
                [],
                {"min": -2147483647, "max": 2147483647},
            ),
            pg_fields.BigIntegerRangeField: (
                fakes.integerrange,
                [],
                {"min": -sys.maxsize, "max": sys.maxsize},
            ),
            pg_fields.FloatRangeField: (fakes.floatrange, [], {}),
            pg_fields.DateTimeRangeField: (fakes.datetimerange, [], {}),
            pg_fields.DateRangeField: (fakes.daterange, [], {}),
            pg_fields.JSONField: (fakes.random_dict, [], {}),
        }
    )


mappings_names = {
    "name": (
        lambda faker, field: field.unique
        and faker.pystr(max_chars=field.max_length or 2700)
        or faker.word()[: field.max_length],
        [],
        {},
    ),  # `name` is too generic to assume it's a person
    "slug": (fakes.slug, [3], {}),
    "first_name": ("first_name", [], {}),
    "last_name": ("last_name", [], {}),
    "full_name": ("full_name", [], {}),
    "email": ("email", [], {}),
    "created": (
        "date_time_between",
        [],
        {"start_date": "-30d", "end_date": "+30d", "tzinfo": TZINFO},
    ),
    "created_at": (
        "date_time_between",
        [],
        {"start_date": "-30d", "end_date": "+30d", "tzinfo": TZINFO},
    ),
    "updated": (
        "date_time_between",
        [],
        {"start_date": "-30d", "end_date": "+30d", "tzinfo": TZINFO},
    ),
    "updated_at": (
        "date_time_between",
        [],
        {"start_date": "-30d", "end_date": "+30d", "tzinfo": TZINFO},
    ),
}



VERSION = "2.2.0"

from django.utils.functional import SimpleLazyObject
from django.utils.module_loading import import_string

from .blueprint import Blueprint
from .lazy import Lazy
from .utils import get_model as M

factory = SimpleLazyObject(lambda: import_string("django_fakery.faker_factory.factory"))



from typing import overload

from django.apps import apps
from django.db import models

from six import string_types

from .types import T


def language_to_locale(language):
    """
    Converts django's `LANGUAGE_CODE` settings to a proper locale code.
    """
    tokens = language.split("-")
    if len(tokens) == 1:
        return tokens[0]
    return "%s_%s" % (tokens[0], tokens[1].upper())


@overload
def get_model(model):
    # type: (str) -> models.Model
    pass


@overload
def get_model(model):
    # type: (T) -> T
    pass


def get_model(model):
    if isinstance(model, string_types):
        model = apps.get_model(*model.split("."))
    return model


def get_model_fields(model):
    fields = list(model._meta._forward_fields_map.items())
    for m2m in model._meta.many_to_many:
        fields.append((m2m.name, m2m))
    return fields


def set_related(instance, attr, value):
    field = getattr(instance, attr)
    field.set(value)



import io
import os

from setuptools import find_packages, setup

here = os.path.abspath(os.path.dirname(__file__))
with io.open(os.path.join(here, "README.rst"), encoding="utf-8") as fp:
    README = fp.read()

with io.open(os.path.join(here, "VERSION")) as version_file:
    VERSION = version_file.read().strip()


setup(
    name="django-fakery",
    version=VERSION,
    url="https://github.com/fcurella/django-fakery/",
    author="Flavio Curella",
    author_email="flavio.curella@gmail.com",
    description="A model instances generator for Django",
    long_description=README,
    license="MIT",
    packages=find_packages(exclude=["docs", "tests", "tests.*"]),
    package_data={"django_fakery": ["py.typed"]},
    platforms=["any"],
    classifiers=[
        "Development Status :: 4 - Beta",
        "Framework :: Django",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Programming Language :: Python",
        "Programming Language :: Python :: 2",
        "Programming Language :: Python :: 2.7",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.4",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
    ],
    entry_points={"pytest11": ["django_fakery = django_fakery.plugin"]},
    install_requires=["Faker>=2.0,<2.1", "Django>=1.11", "six>=1.10.1"],
    extras_require={':python_version<="3.5"': ["typing"]},
    python_requires=">=2.7, !=3.0.*, !=3.1.*, !=3.2.*, !=3.3.*",
    test_suite="tests.runtests.runtests",
)



from datetime import date, datetime, timedelta, tzinfo
from typing import Callable, Optional, Union

from django.conf import settings
from django.utils import timezone

from faker import Faker

ParsableDate = Union[str, int, datetime, timedelta]


def get_timezone():
    # type: () -> Optional[tzinfo]
    return timezone.get_current_timezone() if settings.USE_TZ else None


def future_datetime(end="+30d"):
    # type: (ParsableDate) -> Callable[[int, Faker], datetime]
    """
    Returns a ``datetime`` object in the future (that is, 1 second from now) up
    to the specified ``end``. ``end`` can be a string, another datetime, or a
    timedelta. If it's a string, it must start with `+`, followed by and integer
    and a unit, Eg: ``'+30d'``. Defaults to `'+30d'`

    Valid units are:
    * ``'years'``, ``'y'``
    * ``'weeks'``, ``'w'``
    * ``'days'``, ``'d'``
    * ``'hours'``, ``'hours'``
    * ``'minutes'``, ``'m'``
    * ``'seconds'``, ``'s'``
    """
    return lambda n, f: f.future_datetime(end_date=end, tzinfo=get_timezone())


def future_date(end="+30d"):
    # type: (ParsableDate) -> Callable[[int, Faker], date]
    """
    Returns a ``date`` object in the future (that is, 1 day from now) up to
    the specified ``end``. ``end`` can be a string, another date, or a
    timedelta. If it's a string, it must start with `+`, followed by and integer
    and a unit, Eg: ``'+30d'``. Defaults to `'+30d'`

    Valid units are:
    * ``'years'``, ``'y'``
    * ``'weeks'``, ``'w'``
    * ``'days'``, ``'d'``
    * ``'hours'``, ``'hours'``
    * ``'minutes'``, ``'m'``
    * ``'seconds'``, ``'s'``
    """
    return lambda n, f: f.future_date(end_date=end, tzinfo=get_timezone())


def past_datetime(start="-30d"):
    # type: (ParsableDate) -> Callable[[int, Faker], datetime]
    """
    Returns a ``datetime`` object in the past between 1 second ago and the
    specified ``start``. ``start`` can be a string, another datetime, or a
    timedelta. If it's a string, it must start with `-`, followed by and integer
    and a unit, Eg: ``'-30d'``. Defaults to `'-30d'`

    Valid units are:
    * ``'years'``, ``'y'``
    * ``'weeks'``, ``'w'``
    * ``'days'``, ``'d'``
    * ``'hours'``, ``'h'``
    * ``'minutes'``, ``'m'``
    * ``'seconds'``, ``'s'``
    """
    return lambda n, f: f.past_datetime(start_date=start, tzinfo=get_timezone())


def past_date(start="-30d"):
    # type: (ParsableDate) -> Callable[[int, Faker], date]
    """
    Returns a ``date`` object in the past between 1 day ago and the
    specified ``start``. ``start`` can be a string, another date, or a
    timedelta. If it's a string, it must start with `-`, followed by and integer
    and a unit, Eg: ``'-30d'``. Defaults to `'-30d'`

    Valid units are:
    * ``'years'``, ``'y'``
    * ``'weeks'``, ``'w'``
    * ``'days'``, ``'d'``
    * ``'hours'``, ``'h'``
    * ``'minutes'``, ``'m'``
    * ``'seconds'``, ``'s'``
    """
    return lambda n, f: f.past_date(start_date=start, tzinfo=get_timezone())
